// Counts number of all k-bit strings for small k.
// Orignal written by Michael J. Dinneen
// Modified by Nan Huang
// 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <algorithm>
#include <cassert>
#include <cmath>

#define ALLOFFSETS 0
//#define ALLOFFSETS 1

using namespace std;

struct EOF_Error 
{ 
    int e;
    EOF_Error(int e2) { e = e2; }
};

class BitSeq  // sequence of bits from a raw binary file
{
    private:
	ifstream in;
	
	char byte;	// current byte read from stream
	int offset;  	// last bit 7..0 of byte used

    public:
	BitSeq(const char *fname) : in(fname) { offset=-1; }
	~BitSeq() { in.close(); }
	int eof() { return (in.fail() && in.eof()); }

	char getbit() 
	{
          if (offset==-1) { in.get(byte); offset=7; }
	  return byte & 1<<offset-- ? 1 : 0;
	}

        unsigned int getbits(int cnt)
	{ 
	    unsigned int t = 0;

	    if (offset < cnt)   // can we finish off current byte?
	    {// note: if offset == -1 then t will still be 0			
		t = byte & (1<<offset+1)-1;
		cnt -= offset+1;
		offset=-1; 
	    }

	    // fast version (by bytes)
	    //
            if (cnt>=8) offset=-1;  // set once
	    while (cnt>=8) 
	    { 
		in.get(byte); 
		if (in.eof()) throw EOF_Error(1);
//		assert( !in.eof() );	// more bits in stream?
		t = t*256+((unsigned char) byte); 
		cnt -= 8; 
	    }

	    // slow version (continue by bits)
	    //
	    while (cnt--) { t = t*2+getbit(); }
	    if (in.eof()) throw EOF_Error(2);
	    return t; 
	}
};


int main(int argn, char *argc[])
{
  assert(argn==3);  // input bit filename and bit length to inspect.
    
  unsigned int n = atoi(argc[2]);  // bit length

  cout << "Counting all length n = " << n << " bit strings in file " 
	 << argc[1] << endl;

  unsigned int strs = 1 << n; // number of bit strings to find

  vector<unsigned int> cnts(strs,0);

    //cerr.setf(ios_base::hex, ios_base::basefield); // hex output

#if ALLOFFSETS
  for (int offset=0; offset < n; offset++){
    BitSeq inbits(argc[1]);
    if (offset) inbits.getbits(offset);
    cerr << "Searching file=" << argc[1] << " length=" << n << " offset=" << offset << ' ';
#else
    BitSeq inbits(argc[1]);	     
#endif
    unsigned int cnt = 0;	// counter for number of strings read
    try{
      while (true){
	unsigned int b = inbits.getbits(n); cnt++;
	cnts[b] = cnts[b]+1;
      }
    }
    catch (EOF_Error e){
      cerr << "read " << e.e << '/'<<  cnt << " substrings\n";
    }
#if ALLOFFSETS
  }
#endif
  // code written or modified by Nan starts here
  cout << "Occurance: ";
  for (unsigned int i=0; i < strs; i++)
    cout<<cnts[i]<<' ';
  cout << endl;
  cout << "min: " << *min_element(cnts.begin(), cnts.end()) << endl;//minimum of the count
  cout << "max: " << *max_element(cnts.begin(), cnts.end()) << endl;//maximum of the count
}
