// Rabin/Solovay test using random tests.
// g++ -O2 SS_test1.cpp -lgmpxx -lgmp  -o SS_test1
// 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cassert>

#include <gmpxx.h>

//#include <gmp.h>
//#include <gmp-impl.h>

using namespace std;

struct EOF_Error 
{ 
    int e;
    EOF_Error(int e2) { e = e2; }
};

class BitSeq  // sequence of bits from a raw binary file
{
    private:
	ifstream in;
	
	char byte;	// current byte read from stream
	int offset;  	// last bit 7..0 of byte used

    public:
	BitSeq(const char *fname) : in(fname) { offset=-1; }
	~BitSeq() { in.close(); }
	int eof() { return (in.fail() && in.eof()); }

	char getbit() 
	{
          if (offset==-1) { in.get(byte); offset=7; }
	  return byte & 1<<offset-- ? 1 : 0;
	}

        mpz_class getbits(int cnt)
	{ 
	    mpz_class t(0);

	    if (offset < cnt)   // can we finish off current byte?
	    {// note: if offset == -1 then t will still be 0			
		t = byte & (1<<offset+1)-1;
		cnt -= offset+1;
		offset=-1; 
	    }
#if 1
	    // fast version (by bytes)
	    //
            if (cnt>=8) offset=-1;  // set once
	    while (cnt>=8) 
	    { 
		in.get(byte); 
		if (in.eof()) throw EOF_Error(1);
		t = t*256+((unsigned char) byte); 
		cnt -= 8; 
	    }
#endif
	    // slow version (continue by bits)
	    //
	    while (cnt--) { t = t*2+getbit(); }
	    if (in.eof()) throw EOF_Error(2);
	    return t; 
	}
};

BitSeq *inbits;

bool solovaytest(mpz_class &n)
{
  mpz_class nm1(n), x(1L), y;
  nm1 = nm1 - 1L;

  mpz_class a(0);

  int cnt = mpz_sizeinbase(n.get_mpz_t(),2);
  do {
      a = inbits->getbits(cnt);
      cout << "a=" << a << endl;
  } while (a==0 || a >= n);

  cout << "a=" << a << endl;

  int jacobi = mpz_jacobi(a.get_mpz_t(), n.get_mpz_t());

  if (jacobi==0) return true;

  mpz_class nm1d2(nm1);
  nm1d2 = nm1d2 / 2;
  // test a^(n-1/2) != jacobi (mod n)
  //
  mpz_powm(y.get_mpz_t(), a.get_mpz_t(), nm1d2.get_mpz_t(), n.get_mpz_t());  

  if ((jacobi==-1 && y==nm1) || (jacobi==1 && y==1)) return false;
  return true;
}

int main(int argn, char *argc[])
{
    // A sequence of big int pairs is read from keyboard
    // 
    // Arguments: input bit filename and number of Rabin tests plus
    // an optional read offset from bit file
    assert(argn==3 || argn==4);  

    inbits = new BitSeq(argc[1]);     // random bit source filename
    int k = atoi(argc[2]); 	// number of tests per composite.

    if (argn==4) inbits->getbits(atoi(argc[3])); // skip some bits

    cout << "Trying " << k << " SS tests using " << argc[1] << 
		" as random bit source." << endl;

    int cnt = 0;	// the number of failed tests

    mpz_class p1,p2,n; 	// two primes and their composite product

    int r; mpz_class s,a; // Rabin-Miller test variables for n;
    mpz_class t1,t2;	  // temporary ints

try
{
    while (true)
    {
       cin >> p1;  if (p1==0) break;
       n = p1;

       cout << "n=" << n << endl;

       assert(n % 2 == 1);

       bool composite;
       for (int i=0; i<k; i++)
       {
         composite = solovaytest(n);
	 if (composite) break;
       }
       cout << "my test=" << composite << endl;
    }
}
catch (EOF_Error e)
{
    cerr << "Read past end of file!\n";
}

}

