// Solovay test using random tests (count bits used version).
// g++ -O2 SS_test2.cpp -lgmpxx -lgmp  -o SS_test2.cpp
// 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cassert>

#include <gmpxx.h>

using namespace std;

//#define DOCOMP

struct EOF_Error 
{ 
    int e;
    EOF_Error(int e2) { e = e2; }
};

class BitSeq  // sequence of bits from a raw binary file
{
    private:
	ifstream in;
	
	char byte;	// current byte read from stream
	int offset;  	// last bit 7..0 of byte used
	long bitcnt;

    public:
	BitSeq(const char *fname) : in(fname) { offset=-1; bitcnt=0; }
	~BitSeq() { in.close(); }
	int eof() { return (in.fail() && in.eof()); }

	char getbit() 
	{
	  bitcnt++;
#ifdef DOCOMP
          if (offset==-1) { in.get(byte); byte = ~byte; offset=7; }
#else
          if (offset==-1) { in.get(byte); offset=7; }
#endif
	  return byte & 1<<offset-- ? 1 : 0;
	}

        mpz_class getbits(int cnt)
	{ 
	    bitcnt += cnt;
	    mpz_class t(0);

	    if (offset < cnt)   // can we finish off current byte?
	    {// note: if offset == -1 then t will still be 0			
		t = byte & (1<<offset+1)-1;
		cnt -= offset+1;
		offset=-1; 
	    }

	    // fast version (by bytes)
	    //
            if (cnt>=8) offset=-1;  // set once
	    while (cnt>=8) 
	    { 
		in.get(byte); 
                //cerr << "read byte:" << std::hex << (unsigned int) byte << endl;
#ifdef DOCOMP
		byte= ~byte;
                //cerr << "comp byte:" << std::hex << (unsigned int) byte << endl;
#endif
		if (in.eof()) throw EOF_Error(1);
		t = t*256+((unsigned char) byte); 
		cnt -= 8; 
	    }

	    // slow version (continue by bits)
	    //
	    while (cnt--) { t = t*2+getbit(); }
	    if (in.eof()) throw EOF_Error(2);
	    return t; 
	}

	long count() const { return bitcnt; }
        void ignorebits(int cnt) { bitcnt -= cnt; }
};

BitSeq *inbits;

bool solovaytest(mpz_class &n)
{
  mpz_class nm1(n), x(1L), y;
  nm1 = nm1 - 1L;

  mpz_class a(0);

  int cnt = mpz_sizeinbase(n.get_mpz_t(),2);

  do {
       a = inbits->getbits(cnt);
       if (a==0 || a>= n) inbits->ignorebits(cnt);
       else break;
  } while (true);

  cout << "a=" << a << endl;

  int jacobi = mpz_jacobi(a.get_mpz_t(), n.get_mpz_t());

  if (jacobi==0) return true;

  mpz_class nm1d2(nm1);
  nm1d2 = nm1d2 / 2;

  // test a^(n-1/2) != jacobi (mod n)
  //
  mpz_powm(y.get_mpz_t(), a.get_mpz_t(), nm1d2.get_mpz_t(), n.get_mpz_t());  

  if ((jacobi==-1 && y==nm1) || (jacobi==1 && y==1)) return false;
  return true;
}

int main(int argn, char *argc[])
{
    // A sequence of big int pairs is read from keyboard
    // 
    // Arguments: input bit filename plus an optional read offset from bit file
    assert(argn==2 || argn==3);  

    //BitSeq inbits(argc[1]);     // random bit source filename
    inbits = new BitSeq(argc[1]); // random bit source filename
    int maxk=0;  		  // max number of tests needed per composite.

    if (argn==3) inbits->getbits(atoi(argc[2])); // skip some bits

    cout << "Trying source: " << argc[1] << endl;

    mpz_class n; 	          // input composite product

try
{
    while (true)
    {
       cin >> n;  if (n==0) break;
       cout << "n=" << n << endl;
       assert(n % 2 == 1);

       int i; for (i=1; true; i++)
       {
         if (solovaytest(n)) break;
       }
       if (i>maxk) maxk=i;
       cout << "tests needed " << i << " (max " << maxk << ")" << endl;
    }
}
catch (EOF_Error e)
{
    cerr << "Read past end of file!\n";
}
     cout << "total bits used: " << inbits->count() << endl;
}

